package com.eis.hw;

import java.io.PrintWriter;
import java.util.Iterator;
import java.util.Set;

import javax.resource.ResourceException;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.resource.spi.ValidatingManagedConnectionFactory;
import javax.security.auth.Subject;

/**
 * - parameterloser Konstruktor 
 * - config-properties werden als Bean getter und Setter bereitgestellt 
 * - Es werden mittels equals und hashCode Pools verwaltet:
 * nach 16.4 und 6.5.3.2 der Spec ist eine Implementierung für hashCode
 * und equals auf der Basis der Klasseneigenschaften zu liefern.
 * 
 * Diese Klasse kann {@link ValidatingManagedConnectionFactory}
 * implementieren, der Container wird, wenn entsprechend
 * konfiguriert, regelmäßig alle nichtvaliden ManagedConnection
 * aus dem Pool entfernen.
 */
public final class ManagedConnectionFactoryImpl implements
		ManagedConnectionFactory {

	public final String test = "ManagedConnectionFactoryImpl";

	/**
	 * Die UID der Klasse.
	 */
	private static final long serialVersionUID = 6902822490236740333L;

	/**
	 * Wird mit dem Inhalt der config-property aus der ra.xml belegt.
	 */
	private String hostURL = null;

	/**
	 * Setzt einen Parameter
	 * 
	 * @param hostURL
	 */
	public void setHostURL(String hostURL) {
		this.hostURL = hostURL;
	}

	/**
	 * Liefert einen Parameter
	 * 
	 * @return
	 */
	String getHostURL() {
		return this.hostURL;
	}

	/**
	 * In Containerumgebungen bei der Initialisierung des ResourceAdapters
	 * gerufen (der ConnectionManager wird bereitgestellt)
	 */
	public Object createConnectionFactory(ConnectionManager cm)
			throws ResourceException {
		return new HelloWorldConnectionFactoryImpl(cm, this);
	}

	/**
	 * wird in Standalone Szenarien gerufen, der ResourceAdapter stellt die
	 * Implementierung des ConnectionManager (siehe {@link DefaultConnectionManager}).
	 */
	public Object createConnectionFactory() throws ResourceException {
		return createConnectionFactory(new DefaultConnectionManager());
	}

	/**
	 * Der Container hat sich entschlossen, eine neue physikalische Connection
	 * zu instanziieren (zB keine freie Instanz im Pool).
	 * 
	 * @see javax.resource.spi.ManagedConnectionFactory#createManagedConnection(javax.security.auth.Subject,
	 *      javax.resource.spi.ConnectionRequestInfo)
	 */
	public ManagedConnection createManagedConnection(Subject subject,
			ConnectionRequestInfo info) throws ResourceException {
		if (info == null || info instanceof HelloWorldConnectionRequestInfo) {
			return new ManagedConnectionImpl(this, subject,
					(HelloWorldConnectionRequestInfo) info,
					getLogWriter());
		}
		throw new ResourceException("Unknown ConnectionRequestInfo type: "
				+ info.getClass().getName());
	}

	/**
	 * Der Container versucht mit Hilfe dieses Aufrufs eine geeigeneten Instanz
	 * im Pool zu finden.
	 * 
	 * Der eigentliche Vergleich wird an die 
	 * {@link ManagedConnectionImpl#matches(Subject, ConnectionRequestInfo)} delegiert.
	 * 
	 * Wenn Pooling nicht unterstützt werden kann, muss hier eine
	 * NotSupportedException geworfen werden.
	 */
	public ManagedConnection matchManagedConnections(final Set connectionSet,
			final Subject subject, final ConnectionRequestInfo info)
			throws ResourceException {

		for (final Iterator iter = connectionSet.iterator(); iter.hasNext();) {
			Object obj = iter.next();
			if (obj instanceof ManagedConnectionImpl) {
				ManagedConnectionImpl managedConnection = (ManagedConnectionImpl) obj;
				if (managedConnection.matches(subject, info)) {
					return managedConnection;
				}
			}
		}
		/*
		 * das ist das Signal an den Container, dass nichts gefunden wurde:
		 */
		return null;
	}

	/**
	 * Die Implementierung berücksichtigt alle Konfigurationsparameter dieser
	 * Factory-Instanz - falls in einem Container mehr als einmal der gleiche
	 * ResourceAdapter deployed wird (mit unterschiedlichen Konfigurationen)
	 * wird der Container mit dieser Implementierung seine Pools strukturieren.
	 */
	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + ((hostURL == null) ? 0 : hostURL.hashCode());
		return result;
	}

	/**
	 * Eine equals Implementierung konform zu hashCode
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final ManagedConnectionFactoryImpl other = (ManagedConnectionFactoryImpl) obj;
		if (hostURL == null) {
			if (other.hostURL != null)
				return false;
		} else if (!hostURL.equals(other.hostURL))
			return false;
		return true;
	}

	/**
	 * Der Appserver regelt die Assoziation mit einem Logwriter.
	 */
	private PrintWriter logWriter = null;

	public PrintWriter getLogWriter() throws ResourceException {
		return this.logWriter;
	}

	public void setLogWriter(PrintWriter logWriter) throws ResourceException {
		this.logWriter = logWriter;
	}
}