package com.eis.hw;

import java.io.PrintWriter;
import java.util.LinkedList;
import java.util.List;

import javax.resource.NotSupportedException;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionEvent;
import javax.resource.spi.ConnectionEventListener;
import javax.resource.spi.ConnectionManager;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.LocalTransaction;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionFactory;
import javax.resource.spi.ManagedConnectionMetaData;
import javax.security.auth.Subject;
import javax.transaction.xa.XAResource;

/**
 * Die {@link ManagedConnectionImpl} repräsentiert eine gepoolte, physikalische Verbindung zum EIS und
 * liefert ConnectionHandles.
 * 
 * Jedoch: sämtlicher Code wird im Kontext des Clients ausgeführt! Die ManagedConnection oder ihre
 * ConnectionHandles sind keine entfernten Ressourcen!
 * 
 * ConnectionHandles werden von {@link #getConnection(Subject, ConnectionRequestInfo)} geliefert,
 * eine Methode die jedoch ausschließlich vom ConnectionManager gerufen wird wenn dieser entscheidet,
 * dass neue ConnectionHandles zu erstellen sind. {@link ManagedConnectionImpl} muss die üblichen 
 * Lifecyclemethoden einer vom Container überwachten Ressource anbieten 
 * (Konstruktor, {@link #cleanup()}, {@link #destroy()} und EventListener Methoden).
 * 
 * Instanzen dieser Klasse werden in einem Pool organisiert. Die Implementierung des Pools
 * wird vom Container (als {@link ConnectionManager} bereitgestellt. Wird
 * {@link ConnectionManager#allocateConnection(javax.resource.spi.ManagedConnectionFactory, ConnectionRequestInfo)}
 * grufen nutzt {@link ConnectionManager} die Implementierung von 
 * {@link ManagedConnectionFactory#matchManagedConnections(java.util.Set, Subject, ConnectionRequestInfo)}
 * und diese ruft das hier implementierte {@link #matches(Subject, ConnectionRequestInfo)}.
 * 
 * Wichtig: die Standard {@link #equals(Object)} und {@link #hashCode()} Implementierung darf aber 
 * laut Spec nicht überschrieben werden.
 * 
 * Ein weiterer Vertrag muss Connection Sharing unterstützen. Dabei wird innerhalb einer
 * Transaktion ein und dieselbe Instanz für mehr als ein ConnectionHandle benutzt. 
 * Die Wahrscheinlichkeit für Dead Locks wird dadurch verringert und die Nutzung der Poolinstanzen
 * insgesamt optimiert.
 * 
 * Für Connection Sharing ruft der Container {@link #associateConnection(Object)} in einer von
 * ihm bestimmten Reihenfolge. ConnectionHandles die vorübergehend NICHT mit ihrer {@link ManagedConnectionImpl}
 * Instanz assoziiert sind, dürfen diese Instanz auch nicht benutzen. Versuchen sie es dennoch,
 * wird eine {@link ResourceException} geworfen.
 * 
 * Ein weiterer Vertrag betrifft das Transaktionsmanagement, das hier nicht ausimplementiert ist.
 * 
 * Die Spec fordert, dass für kommende Erweiterungen alle Implementierung die von den
 * ConnectionHandles gerufen werden, nebenläufig implementiert werden.
 * 
 * 
 */
final class ManagedConnectionImpl implements ManagedConnection {

	
	/**
	 * Für die Arbeit mit dem EIS sind oft Credentials (user, password)
	 * notwendig, die mit diesem Subject weitergereicht werden.
	 * 
	 * Die Subject-Instanz wird vom Container bereitgestellt. 
	 * SSO Funktionalität kann hiermit realisiert werden.
	 * 
	 * @see Subject
	 */
	private final Subject subject;

	/**
	 * Für die Arbeit mit dem EIS sind oft Client-Parameter bei
	 * der Etablierung der EIS Connection nötig. Diese werden
	 * hier abgelegt.
	 */
	private final HelloWorldConnectionRequestInfo info;

	/**
	 * ManagedConnection - Konstruktor.
	 * 
	 * Hier erfolgt die physikalische Connection zum EIS. Sie
	 * wird solange gehalten und belegt diese Ressourcen, bis
	 * destroy gerufen wird.
	 * 
	 * @param mcf Die ManagedConnectionFactory Instanz, hier können
	 * die Konfigurationsparameter des Deployments abgefragt werden.
	 * @param subject Die User Credentials
	 * @param info Die ConnectionRequestInfo
	 * @param defaultLogWriter der LogWriter der ManagedConenctionFactory
	 */
	ManagedConnectionImpl(final ManagedConnectionFactoryImpl mcf,
			final Subject subject, 
			final HelloWorldConnectionRequestInfo info,
			final PrintWriter defaultLogWriter) {
		// Subject und info legen die Identität dieser Instanz fest:
		this.subject = subject;
		this.info = info;
		this.logWriter = defaultLogWriter;
		/*
		 * hier Code, der physikalische Ressource zur Arbeit
		 * mit dem EIS allociert. Diese Resourcen werden bis
		 * zum Aufruf von destroy gehalten. Zur Etablierung der
		 * Ressourcen werden potenziell auch die Konfigurations-
		 * parameter der ManagedConnectionFactory verwendet:
		 */ 
		String hostURL = mcf.getHostURL();
		/*
		 * Jetzt kann die physikalische Connection aufgebaut werden.
		 * Falls dazu ein Username aus dem Subject zu holen ist, dann
		 * in getUserName implementieren - dort wird es auch von
		 * ManagedConnectionMetaDataImpl benutzt.
		 */
		this.logWriter.print(this + " created: ");
		this.logWriter.print("Host URL is: " + hostURL);
	}

	/**
	 * Das momentan mit dieser {@link ManagedConnection} assoziierte ConnectionHandle 
	 */
	private HelloWorldConnectionImpl associatedHandle = null; 

	/**
	 * Der Server kann im Rahmen des Connection Sharings sich jederzeit die 
	 * Assoziation eines Connection Handles zu dieser Managed Connection verbiegen 
	 * und ruft zu diesem Zweck diese Methode.
	 */
	public void associateConnection(Object connection) throws ResourceException {
		disassociateConnection();
		HelloWorldConnectionImpl connectionHandle = (HelloWorldConnectionImpl) connection;
		connectionHandle.setManagedConnetion(this);
		this.associatedHandle = connectionHandle;
	}
	
	/**
	 * Kappt die Verbindung zwischen dieser Instanz und dem 
	 * gerade assoziierten ConnectionHandle.
	 */
	private void disassociateConnection() {
		if (this.associatedHandle != null) {
			this.associatedHandle.setManagedConnetion(null);
			this.associatedHandle = null;
		}		
	}

	/**
	 * Der Container entschließt sich, eine ManagedConnection aus dem Pool zu
	 * entfernen und ruft diese Methode damit alle physikalischen Resourcen
	 * freigegeben werden können.
	 * 
	 * Wird regulär gerufen, wenn die konfigurierte max-idle Zeit
	 * überschritten wird.
	 */
	public void destroy() throws ResourceException {
		this.logWriter.print(this + " destroy");
		// zunächst ein cleanup
		cleanup();
		// dann Code, der alle physikalischen Ressourcen freigibt 
	}

	/**
	 * Wird ausschließlich vom Container gerufen, diese Instanz wird 
	 * gleich in den Pool befördert, der Zustand der physikalischen 
	 * Connection muss intakt bleiben.
	 * 
	 * Wird regulär gerufen, wenn das assoziierten ConnectionHandle
	 * die Connection schließt (und dieses Signal ordnungsgemäß
	 * an den Container weitergereicht wird, siehe
	 * {@link #closeHandle(HelloWorldConnection)}.
	 */
	public void cleanup() throws ResourceException {
		this.logWriter.print(this + " cleanup");
		// löse die Assoziation wenn nicht schon geschehen:
		disassociateConnection();
	}

	/**
	 * Wird im Zusammenhang mit {@link ConnectionManager#allocateConnection(ManagedConnectionFactory, ConnectionRequestInfo)}
	 * gerufen und liefert ein Connection Handle das mit dieser Instanz assoziiert ist.
	 * 
	 * Der Container kann im Rahmen des Connection Sharings mehr als einmal
	 * diese Methode aufrufen, auch wenn die das erste ConnectionHandle noch
	 * nicht die close-Methode gerufen hat.
	 */
	public Object getConnection(final Subject subject,
			final ConnectionRequestInfo connectionRequestInfo)
			throws ResourceException {

		this.logWriter.print(this + " getConnection");

		HelloWorldConnection connection = new HelloWorldConnectionImpl();
		associateConnection(connection);
		/* connectionRequestInfo und subject können hier bei Bedarf
		 * für Reconnect etc. verwendet werden (Auffrischen der
		 * physikalischen Connection) 
		 */ 
		return connection;
	}

	public XAResource getXAResource() throws ResourceException {
		throw new NotSupportedException("XA transaction not supported");
	}

	public LocalTransaction getLocalTransaction() throws ResourceException {
		throw new NotSupportedException("Local transaction not supported");
	}

	/**
	 * Der Default LogWriter wird von der ManagedConectionFactoryImpl
	 * bereitgestellt - er kann aber vom Server auch gesetzt werden...
	 */
	private PrintWriter logWriter = null;

	public PrintWriter getLogWriter() throws ResourceException {
		return this.logWriter;
	}

	public void setLogWriter(PrintWriter logWriter) throws ResourceException {
		this.logWriter = logWriter;
	}

	public ManagedConnectionMetaData getMetaData() throws ResourceException {
		return new ManagedConnectionMetaDataImpl(this);
	}

	/**
	 * Implementierung der EIS Logik.
	 * 
	 * Diese Implementierung erfolgt mittels der im Konstruktor
	 * akquirierten EIS Ressourcen und muss potentiell auch nebenläufig
	 * funktionieren.
	 * 
	 * @return Ein EIS Ergebnis
	 */
	String doHelloWorldCall() {
		
		if (this.info == null) {
			return "Hello World";
		}
		String message = this.info.getMessage();
		// So kann man einen Fehlerverhalten testen
		if ("ERROR".equals(message)) {
			errorOccured();
		}
		return "Hello " + message + "!";		
	}
	
	/**
	 * Ein Connection Handle signalisiert, dass es geschlossen wird.
	 *
	 * Der Container hat sich als Listener angemeldet und wird unter 
	 * zuhilfenahme dieses Events seine Pools verwalten. 
	 * 
	 * @param connection
	 */
	void closeHandle(HelloWorldConnection connection) {
		this.logWriter.print(this + " closeHandle: ");
		disassociateConnection();
		for (ConnectionEventListener listener : this.listenerList) {
			ConnectionEvent event = new ConnectionEvent(this,
					ConnectionEvent.CONNECTION_CLOSED);
			event.setConnectionHandle(connection);
			listener.connectionClosed(event);
		}
	}

	/**
	 * Alle Listener (der Container!) werden benachrichtigt, dass ein Fehler passiert ist. 
	 */
	private void errorOccured() {
		this.logWriter.print(this + " error: ");
		for (ConnectionEventListener listener : this.listenerList) {
			ConnectionEvent event = new ConnectionEvent(this,
					ConnectionEvent.CONNECTION_ERROR_OCCURRED);
			listener.connectionClosed(event);
		}
	}

	private final List<ConnectionEventListener> listenerList = new LinkedList<ConnectionEventListener>();
	
	public void addConnectionEventListener(ConnectionEventListener listener) {
		this.listenerList.add(listener);
	}
	
	public void removeConnectionEventListener(ConnectionEventListener listener) {
		this.listenerList.remove(listener);
	}

	/**
	 * Beispielhafte Implementierung zur Unterstützung von 
	 * {@link ManagedConnectionFactory#matchManagedConnections}.
	 * 
	 * this.subject und this.info charakterisieren die Identität dieser
	 * ManagedConnection Instanz - zur Implementierung von matches
	 * sollten aber nur diejenigen Informationen verwendet werden,
	 * die tatsächlich bei der Etablierung der physikalischen
	 * Verbindung zum EIS gebraucht werden.
	 * 
	 * @param subject
	 * @param info
	 * @return 
	 */
	boolean matches(final Subject subject, final ConnectionRequestInfo info) {
		this.logWriter.print(this + " matches: ");
		if (info == null) {
			if (this.info != null) {
				return false;
			}
		} else if (!info.equals(this.info)) {
			return false;
		}
		if (subject == null) {
			if (this.subject != null) {
				return false;
			}
		} else if (!subject.equals(this.subject)) {
			return false;
		}
		return true;
	}
	
	/**
	 * Ermittelt aus dem Subject den Usernamen.
	 * 
	 * @return Der Username für den Aufbau der Verbindung.
	 */
	String getUserName() {
		if (this.subject != null) {
		  this.subject.getPrincipals();
 		  //... mehr Code
		}
		return null; // dummy
	}

	/**
	 * In Section 16.4 der Spec ist gefordert, dass die Standard equals 
	 * Implementierung NICHT überschrieben wird ("no two Java objects are 
	 * considerd equals").
	 * 
	 * {@link #matches(Subject, ConnectionRequestInfo)}.
	 */ 	
	@Override
	public final boolean equals(Object obj) {
		// this implementation must not be changed
		return super.equals(obj);
	}
	
	/**
	 * In Section 16.4 der Spec ist gefordert, dass die Standard equals 
	 * Implementierung NICHT überschrieben wird ("no two Java objects are 
	 * considerd equals"), damit muss auch {@link #hashCode()} bleiben.
	 */ 	
	@Override
	public final int hashCode() {
		// this implementation must not be changed
		return super.hashCode();
	}
	
}